from typing import Dict, List, Optional, Sequence

import json
from abc import ABC, abstractmethod

import numpy as np
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential

from ..config import LLMConfigurationError
from ..errors import LLMGenerationError, LLMImportError
from . import LLMClient, LLMFunctionCall, LLMLogger, LLMMessage
from .base import LLMToolCall

try:
    import openai
except ImportError as err:
    raise LLMImportError(flavor="llm") from err

AUTH_ERROR_MESSAGE = (
    "Could not authenticate with OpenAI API. Please make sure you have configured the API key by "
    "setting OPENAI_API_KEY in the environment."
)


class BaseOpenAIClient(LLMClient, ABC):
    _max_embedding_chunk_size = 2048

    def __init__(self, model: str):
        self._logger = LLMLogger()
        self.model = model

    @property
    def logger(self) -> LLMLogger:
        return self._logger

    @abstractmethod
    def _completion(
        self,
        messages: Sequence,
        functions: Sequence = None,
        temperature: float = 1.0,
        function_call: Optional[Dict] = None,
        max_tokens=None,
        caller_id: Optional[str] = None,
        tools=None,
        tool_choice=None,
        seed: Optional[int] = None,
    ) -> dict:
        ...

    @staticmethod
    def _serialize_function_call(function_call: LLMFunctionCall) -> Dict:
        return {"name": function_call.name, "arguments": json.dumps(function_call.arguments)}

    @staticmethod
    def _serialize_tool_call(tool_call: LLMToolCall) -> Dict:
        return {
            "id": tool_call.id,
            "type": tool_call.type,
            "function": BaseOpenAIClient._serialize_function_call(tool_call.function),
        }

    @staticmethod
    def _serialize_tool_calls(tool_calls: List[LLMToolCall]) -> List[Dict]:
        return [BaseOpenAIClient._serialize_tool_call(tool_call) for tool_call in tool_calls]

    @staticmethod
    def _serialize_message(response: LLMMessage) -> Dict:
        result = {
            "role": response.role,
            "content": response.content,
            "function_call": (
                BaseOpenAIClient._serialize_function_call(response.function_call) if response.function_call else None
            ),
            "tool_calls": BaseOpenAIClient._serialize_tool_calls(response.tool_calls) if response.tool_calls else None,
        }

        return {key: value for key, value in result.items() if value is not None}

    @staticmethod
    def _parse_function_call(function_call) -> LLMFunctionCall:
        try:
            return LLMFunctionCall(
                name=function_call["name"],
                arguments=json.loads(function_call["arguments"]),
            )
        except (json.JSONDecodeError, KeyError) as err:
            raise LLMGenerationError("Could not parse function call") from err

    @staticmethod
    def _parse_tool_call(tool_call) -> LLMToolCall:
        return LLMToolCall(
            id=tool_call["id"],
            type=tool_call["type"],
            function=BaseOpenAIClient._parse_function_call(tool_call["function"]),
        )

    @staticmethod
    def _parse_tool_calls(tool_calls) -> List[LLMToolCall]:
        return [BaseOpenAIClient._parse_tool_call(tool_call) for tool_call in tool_calls]

    @staticmethod
    def _parse_message(response) -> LLMMessage:
        return LLMMessage(
            role=response["role"],
            content=response["content"],
            function_call=(
                BaseOpenAIClient._parse_function_call(response["function_call"])
                if "function_call" in response and response["function_call"] is not None
                else None
            ),
            tool_calls=(
                BaseOpenAIClient._parse_tool_calls(response["tool_calls"])
                if "tool_calls" in response and response["tool_calls"] is not None
                else None
            ),
        )

    def complete(
        self,
        messages: Sequence[LLMMessage],
        functions=None,
        temperature=0.5,
        max_tokens=None,
        function_call: Optional[Dict] = None,
        caller_id: Optional[str] = None,
        tools=None,
        tool_choice=None,
        seed: Optional[int] = None,
    ):
        llm_message = self._completion(
            messages=[
                BaseOpenAIClient._serialize_message(message) if isinstance(message, LLMMessage) else message
                for message in messages
            ],
            temperature=temperature,
            functions=functions,
            function_call=function_call,
            max_tokens=max_tokens,
            caller_id=caller_id,
            tools=tools,
            tool_choice=tool_choice,
            seed=seed,
        )

        return BaseOpenAIClient._parse_message(llm_message)

    @abstractmethod
    def _embeddings_generation(self, texts: Sequence[str], model: str):
        ...

    def embeddings(
        self, texts: Sequence[str], model: str = "text-embedding-ada-002", chunk_size: int = 2048
    ) -> np.ndarray:
        chunks_indices = range(chunk_size, len(texts), chunk_size)
        chunks = np.split(texts, chunks_indices)
        embedded_chunks = [self._embeddings_generation(chunk, model) for chunk in chunks]
        return np.stack([emb for embeddings in embedded_chunks for emb in embeddings])


class LegacyOpenAIClient(BaseOpenAIClient):
    """OpenAI client for versions <= 0.28.1"""

    def __init__(self, model: str, openai_api_key=None, openai_organization=None):
        self.openai_api_key = openai_api_key
        self.openai_organization = openai_organization
        super().__init__(model)

    @retry(retry=retry_if_exception_type(openai.OpenAIError), stop=stop_after_attempt(3), wait=wait_exponential(3))
    def _completion(
        self,
        messages: Sequence,
        functions: Sequence = None,
        temperature: float = 1.0,
        function_call: Optional[Dict] = None,
        max_tokens=None,
        caller_id: Optional[str] = None,
        tools=None,
        tool_choice=None,
        seed: Optional[int] = None,
    ):
        extra_params = dict()
        if function_call is not None:
            extra_params["function_call"] = function_call
        if functions is not None:
            extra_params["functions"] = functions
        if tools is not None:
            extra_params["tools"] = tools
        if tool_choice is not None:
            extra_params["tool_choice"] = tool_choice
        if seed is not None:
            extra_params["seed"] = seed

        try:
            completion = openai.ChatCompletion.create(
                model=self.model,
                messages=[
                    BaseOpenAIClient._serialize_message(message) if isinstance(message, LLMMessage) else message
                    for message in messages
                ],
                temperature=temperature,
                max_tokens=max_tokens,
                **extra_params,
                api_key=self.openai_api_key,
                organization=self.openai_organization,
            )
        except openai.error.AuthenticationError as err:
            raise LLMConfigurationError(AUTH_ERROR_MESSAGE) from err

        self._logger.log_call(
            prompt_tokens=completion["usage"]["prompt_tokens"],
            sampled_tokens=completion["usage"]["completion_tokens"],
            model=self.model,
            client_class=self.__class__.__name__,
            caller_id=caller_id,
        )

        return completion["choices"][0]["message"]

    def _embeddings_generation(self, texts: Sequence[str], model: str):
        try:
            out = openai.Embedding.create(input=list(texts), engine=model)
            embeddings = [element["embedding"] for element in out["data"]]
        except openai.error.InvalidRequestError as err:
            raise ValueError(
                f"The embedding model: '{model}' was not found,"
                "make sure the model is correctly deployed on your endpoint."
            ) from err
        except Exception as err:
            raise RuntimeError("Embedding creation failed.") from err

        return embeddings


class OpenAIClient(BaseOpenAIClient):
    def __init__(self, model: str, client=None):
        self._client = client or openai.OpenAI()

        super().__init__(model)

    def _completion(
        self,
        messages: Sequence,
        functions: Sequence = None,
        temperature: float = 1.0,
        function_call: Optional[Dict] = None,
        max_tokens=None,
        caller_id: Optional[str] = None,
        tools=None,
        tool_choice=None,
        seed=None,
    ):
        extra_params = dict()
        if function_call is not None:
            extra_params["function_call"] = function_call
        if functions is not None:
            extra_params["functions"] = functions
        if tools is not None:
            extra_params["tools"] = tools
        if tool_choice is not None:
            extra_params["tool_choice"] = tool_choice
        if seed is not None:
            extra_params["seed"] = seed

        try:
            completion = self._client.chat.completions.create(
                model=self.model,
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                **extra_params,
            )
        except openai.AuthenticationError as err:
            raise LLMConfigurationError(AUTH_ERROR_MESSAGE) from err

        self._logger.log_call(
            prompt_tokens=completion.usage.prompt_tokens,
            sampled_tokens=completion.usage.completion_tokens,
            model=self.model,
            client_class=self.__class__.__name__,
            caller_id=caller_id,
        )

        return completion.choices[0].message.model_dump()

    def _embeddings_generation(self, texts: Sequence[str], model: str):
        try:
            out = self._client.embeddings.create(input=list(texts), model=model)
            embeddings = [element.embedding for element in out.data]
        except openai.NotFoundError as err:
            raise ValueError(
                f"The embedding model: '{model}' was not found,"
                "make sure the model is correctly deployed on "
                f"the specified endpoint: {self._client._base_url}."
            ) from err
        except Exception as err:
            raise RuntimeError("Embedding creation failed.") from err

        return embeddings



from typing import Any, Literal

from enum import Enum
from uuid import UUID

import pydantic
from packaging import version

from giskard.core.validation import ConfiguredBaseModel

IS_PYDANTIC_V2 = version.parse(pydantic.version.VERSION) >= version.parse("2.0")


# Make it so we can use enum in pydantic to parse object
class MLWorkerAction(str, Enum):
    getInfo = "getInfo"
    runAdHocTest = "runAdHocTest"
    datasetProcessing = "datasetProcessing"
    runTestSuite = "runTestSuite"
    runModel = "runModel"
    runModelForDataFrame = "runModelForDataFrame"
    explain = "explain"
    explainText = "explainText"
    echo = "echo"
    generateTestSuite = "generateTestSuite"
    stopWorker = "stopWorker"
    getCatalog = "getCatalog"
    generateQueryBasedSlicingFunction = "generateQueryBasedSlicingFunction"
    getPush = "getPush"
    createSubDataset = "createSubDataset"
    abort = "abort"
    getLogs = "getLogs"
    createDataset = "createDataset"

    @classmethod
    def __get_validators__(cls):
        if IS_PYDANTIC_V2:
            yield cls.validate
        else:
            yield cls.validate_v1

    @classmethod
    def validate(cls, v, *args, **kwargs):
        return cls.validate_v1(v)

    @classmethod
    def validate_v1(cls, value, values=None, config=None, field=None):
        try:
            return cls[value]
        except KeyError as e:
            raise ValueError(f"Unknown value {value}") from e


class ActionPayload(ConfiguredBaseModel):
    id: UUID
    action: MLWorkerAction
    param: Any


class ConfigPayload(ConfiguredBaseModel):
    config: Literal["MAX_STOMP_ML_WORKER_REPLY_SIZE"]
    value: int



from typing import Callable, Optional, Sequence

import numpy as np
import pandas as pd

from ..core.errors import GiskardInstallationError


class Document:
    """A class to wrap the elements of the knowledge base into a unified format."""

    def __init__(self, document: dict, features: Optional[Sequence] = None):
        features = features if features is not None else list(document.keys())

        if len(features) == 1:
            self.content = document[features[0]]
        else:
            self.content = "\n".join(f"{feat}: {document[feat]}" for feat in features)

        self.metadata = document


class VectorStore:
    """Stores all embedded Document of the knowledge base.
    Relies on `FlatIndexL2` class from FAISS.
    """

    def __init__(self, documents: Sequence[Document], embeddings: np.array, embedding_fn: Callable):
        if len(embeddings) == 0 or len(documents) == 0:
            raise ValueError("Documents and embeddings must contains at least one element.")
        if len(embeddings) != len(documents):
            raise ValueError("Documents and embeddings must have the same length.")

        try:
            from faiss import IndexFlatL2
        except ImportError as err:
            raise GiskardInstallationError(flavor="llm") from err

        self.embeddings = embeddings
        self.documents = documents
        self.embedding_fn = embedding_fn

        self.dimension = self.embeddings[0].shape[0]
        self.index = IndexFlatL2(self.dimension)
        self.index.add(self.embeddings)

    @classmethod
    def from_df(cls, df: pd.DataFrame, embedding_fn: Callable, features: Sequence[str] = None):
        if len(df) > 0:
            documents = [Document(knowledge_chunk, features=features) for knowledge_chunk in df.to_dict("records")]
            raw_texts = [d.content for d in documents]
            embeddings = embedding_fn(raw_texts).astype("float32")
            return cls(documents, embeddings, embedding_fn)
        else:
            raise ValueError("Cannot generate a vector store from empty DataFrame.")

    def similarity_search_with_score(self, query: Sequence[str], k: int) -> Sequence:
        query_emb = self.embedding_fn(query).astype("float32")
        return self.vector_similarity_search_with_score(query_emb, k)

    def vector_similarity_search_with_score(self, query_emb: np.ndarray, k: int) -> Sequence:
        query_emb = np.atleast_2d(query_emb)
        distances, indices = self.index.search(query_emb, k)
        return [(self.documents[i], d) for d, i in zip(distances[0], indices[0])]



import numpy as np

from ....core.test_result import TestResult, TestResultDetails, TestResultStatus
from ....datasets.base import Dataset
from ....llm import LLMImportError
from ....llm.evaluators import PerRowRequirementEvaluator
from ....models.base import BaseModel
from ....registry.decorators import test
from .. import debug_description_prefix
from .output_requirements import _test_output_against_requirement


@test(
    name="Evaluation of model output exact match to the ground truth ",
    tags=["llm", "ground-truth"],
    debug_description=debug_description_prefix + "that are <b>generating result differing from ground truth</b>.",
)
def test_llm_ground_truth(model: BaseModel, dataset: Dataset, threshold: float = 0.5):
    if dataset.target is None:
        raise ValueError(f"Provided dataset ({dataset}) does not have any ground truth (target)")

    pred = model.predict(dataset)

    passed = np.array(pred.prediction) == dataset.df[dataset.target]
    metric = len([p for p in passed if p]) / len(passed)

    return TestResult(
        passed=metric >= threshold,
        metric=metric,
        details=TestResultDetails(
            inputs=dataset.df.loc[:, model.meta.feature_names].to_dict("list"),
            outputs=list(pred.prediction),
            results=[TestResultStatus.PASSED if result else TestResultStatus.FAILED for result in passed],
            metadata={"target": list(dataset.df[dataset.target])},
        ),
    )


@test(
    name="Evaluation of model output similarity to the ground truth",
    tags=["llm", "ground-truth"],
    debug_description=debug_description_prefix + "that are <b>generating result differing from ground truth</b>.",
)
def test_llm_ground_truth_similarity(
    model: BaseModel, dataset: Dataset, output_sensitivity: float = 0.15, threshold: float = 0.5, idf: bool = False
):
    if dataset.target is None:
        raise ValueError(f"Provided dataset ({dataset}) does not have any ground truth (target)")

    pred = model.predict(dataset)

    try:
        import evaluate
    except ImportError as err:
        raise LLMImportError() from err

    scorer = evaluate.load("bertscore")
    score = scorer.compute(
        predictions=pred.prediction,
        references=dataset.df[dataset.target],
        model_type="distilbert-base-multilingual-cased",
        idf=idf,
    )
    passed = np.array(score["f1"]) > 1 - output_sensitivity
    metric = len([p for p in passed if p]) / len(passed)

    return TestResult(
        passed=metric >= threshold,
        metric=metric,
        details=TestResultDetails(
            inputs=dataset.df.loc[:, model.meta.feature_names].to_dict("list"),
            outputs=list(pred.prediction),
            results=[TestResultStatus.PASSED if result else TestResultStatus.FAILED for result in passed],
            metadata={"target": list(dataset.df[dataset.target]), "F1 similarity": score["f1"]},
        ),
    )


@test(
    name="Per row evaluation of model output using an LLM (LLM-as-a-judge)",
    tags=["llm", "llm-as-a-judge"],
    debug_description=debug_description_prefix + "that are <b>failing the evaluation criteria</b>.",
)
def test_llm_as_a_judge_ground_truth_similarity(
    model: BaseModel, dataset: Dataset, prefix: str = "The requirement should be similar to: ", rng_seed: int = 1729
):
    """Evaluates the model output against its ground truth  with another LLM (LLM-as-a-judge).

    The model outputs over a given dataset will be validated against the
    dataset target using GPT-4 (note that this requires you to set the
    `OPENAI_API_TOKEN` environment variable for the test to run correctly).

    Parameters
    ----------
    model : BaseModel
        The generative model to test.
    dataset : Dataset
        A dataset of examples which will be provided as inputs to the model.
    prefix : str
        The prefix instructing how the answer should be according to the ground truth”.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if dataset.target is None:
        raise ValueError(f"Provided dataset ({dataset}) does not have any ground truth (target)")

    return _test_output_against_requirement(
        model, dataset, PerRowRequirementEvaluator(dataset.df.loc[:, [dataset.target]], prefix, rng_seed=rng_seed)
    )



from typing import List, Tuple

import logging
import re
import string
from dataclasses import dataclass

from ...core.test_result import TestResultStatus, create_test_result_details
from ...datasets.base import Dataset
from ...models.base.model import BaseModel
from ..errors import LLMGenerationError
from .base import BaseEvaluator, EvaluationResult

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@dataclass(frozen=True)
class StringMatcherConfig:
    expected_strings: Tuple[str]
    all_expected_strings_must_be_found: bool = True
    exact_matching: bool = False
    word_matching: bool = False
    case_sensitive: bool = True
    punctuation_sensitive: bool = True
    evaluation_method_name: str = "StringMatchingMethod"


class StringMatcher:
    def __init__(self, config: StringMatcherConfig) -> None:
        self.config = config

    def normalize_text(self, text):
        if not self.config.case_sensitive:
            text = text.lower()
        if not self.config.punctuation_sensitive:
            text = text.translate(str.maketrans("", "", string.punctuation))
        return text

    def evaluate_single_string(self, string: str, text: str):
        n_string = self.normalize_text(string)
        n_text = self.normalize_text(text)
        if self.config.exact_matching:
            return n_string == n_text
        if self.config.word_matching:
            return re.search(r"\b" + re.escape(n_string) + r"\b", text) is not None
        return n_string in n_text

    def evaluate(self, text: str):
        matches = (self.evaluate_single_string(string, text) for string in self.config.expected_strings)
        if self.config.all_expected_strings_must_be_found:
            return all(matches)
        return any(matches)


class StringMatcherEvaluator(BaseEvaluator):
    def evaluate(self, model: BaseModel, dataset: Dataset, evaluator_configs: List[StringMatcherConfig]):
        succeeded = []
        failed = []
        errored = []
        model_inputs = dataset.df.loc[:, model.feature_names].to_dict("records")
        model_outputs = model.predict(dataset).prediction
        status = []

        for idx, inputs, outputs, config in zip(dataset.df.index, model_inputs, model_outputs, evaluator_configs):
            string_matcher = StringMatcher(config)

            try:
                injection_success = string_matcher.evaluate(outputs)
            except LLMGenerationError as err:
                status.append(TestResultStatus.ERROR)
                errored.append({"message": str(err), "sample": inputs})
                continue

            if not injection_success:
                status.append(TestResultStatus.PASSED)
                succeeded.append({"input_vars": inputs, "model_output": outputs})
            else:
                status.append(TestResultStatus.FAILED)
                failed.append({"input_vars": inputs, "model_output": outputs})

        return EvaluationResult(
            failure_examples=failed,
            success_examples=succeeded,
            errors=errored,
            details=create_test_result_details(dataset, model, model_outputs, status),
        )



from ...core.test_result import TestResultStatus, create_test_result_details
from ...datasets import Dataset
from ...models.base.model import BaseModel
from ..client.base import LLMMessage
from ..errors import LLMGenerationError
from .base import EVALUATE_MODEL_FUNCTIONS, EvaluationResult, LLMBasedEvaluator

CORRECTNESS_EVALUATION_PROMPT = """Your role is to test AI models. Your task consists in assessing whether a model output correctly answers a question. 
You are provided with the ground truth answer to the question. Your task is then to evaluate if the model answer is close to the ground thruth answer. 

You are auditing the following model:

Model name: {model_name}
Model description: {model_description}

Here is the question that was asked to the model and its output, followed by the expected ground truth answer:

QUESTION:
###
{question}
###

MODEL OUTPUT:
###
{model_output}
###

GROUND TRUTH:
###
{ground_truth}
###

Think step by step and consider the model output in its entirety. Remember: you need to have a strong and sound reason to support your evaluation.
Call the `evaluate_model` function with the result of your evaluation.
"""


class CorrectnessEvaluator(LLMBasedEvaluator):
    """Assess the correctness of a model answers given questions and associated reference answers."""

    _default_eval_prompt = CORRECTNESS_EVALUATION_PROMPT

    def _make_evaluate_functions(self):
        return EVALUATE_MODEL_FUNCTIONS

    def _make_evaluate_prompt(self, model_name, model_description, question, model_output, ground_truth):
        return self.eval_prompt.format(
            model_name=model_name,
            model_description=model_description,
            question=question,
            model_output=model_output,
            ground_truth=ground_truth,
        )

    def evaluate(
        self,
        model: BaseModel,
        dataset: Dataset,
        question_col: str = "question",
        reference_answer_col: str = "reference_answer",
    ):
        if not (question_col in dataset.df and reference_answer_col in dataset.df):
            raise ValueError(
                f"Missing required columns in the evaluation dataset. Make sure the dataset has columns {question_col} and {reference_answer_col}."
            )

        if question_col not in model.feature_names:
            raise ValueError(
                f"Model has no feature '{question_col}'. Make sure your Model wrapper accepts '{question_col}'."
            )

        model_outputs = model.predict(dataset).prediction

        succeeded = []
        failed = []
        errored = []
        status = []
        reasons = []
        for evaluation_question, model_output in zip(dataset.df.to_dict("records"), model_outputs):
            try:
                passed, reason = self._evaluate_single(
                    model,
                    evaluation_question[question_col],
                    evaluation_question[reference_answer_col],
                    model_output,
                )
                reasons.append(reason)
                sample = {
                    **evaluation_question,
                    "reason": reason,
                    "model_output": model_output,
                    "model_evaluation": passed,
                }
                if passed:
                    succeeded.append(sample)
                    status.append(TestResultStatus.PASSED)
                else:
                    failed.append(sample)
                    status.append(TestResultStatus.FAILED)
            except LLMGenerationError as err:
                errored.append({"message": str(err), "sample": {**evaluation_question, "model_output": model_output}})
                reasons.append(str(err))
                status.append(TestResultStatus.ERROR)

        return EvaluationResult(
            failure_examples=failed,
            success_examples=succeeded,
            errors=errored,
            details=create_test_result_details(dataset, model, model_outputs, status, {"reason": reasons}),
        )

    def _evaluate_single(self, model: BaseModel, question, reference_answer, model_output):
        prompt = self._make_evaluate_prompt(
            model.meta.name,
            model.meta.description,
            question,
            model_output,
            reference_answer,
        )

        out = self.llm_client.complete(
            [LLMMessage(role="system", content=prompt)],
            tools=self._make_evaluate_functions(),
            tool_choice={"type": "function", "function": {"name": "evaluate_model"}},
            temperature=self.llm_temperature,
            caller_id=self.__class__.__name__,
            seed=self.rng_seed,
        )

        try:
            passed_test = out.tool_calls[0].function.arguments["passed_test"]
            reason = out.tool_calls[0].function.arguments.get("reason")
        except (AttributeError, KeyError, IndexError):
            raise LLMGenerationError("Invalid function call arguments received")

        return passed_test, reason



from typing import Any, Dict, Optional, Sequence

from abc import ABC, abstractmethod
from dataclasses import dataclass

from ...core.test_result import TestResultDetails, TestResultStatus
from ...datasets.base import Dataset
from ...models.base.model import BaseModel
from ..client import LLMClient, get_default_client
from ..errors import LLMGenerationError

EVALUATE_MODEL_FUNCTIONS = [
    {
        "type": "function",
        "function": {
            "name": "evaluate_model",
            "description": "Evaluates if the model passes the test",
            "parameters": {
                "type": "object",
                "properties": {
                    "passed_test": {
                        "type": "boolean",
                        "description": "true if the model successfully passes the test",
                    },
                    "reason": {
                        "type": "string",
                        "description": "optional short description of why the model does not pass the test, in 1 or 2 short sentences",
                    },
                },
                "required": ["passed_test"],
            },
        },
    },
]


@dataclass
class EvaluationResult:
    failure_examples: Sequence[dict]
    success_examples: Sequence[dict]
    errors: Sequence[dict]
    details: Optional[TestResultDetails] = None

    @property
    def passed(self):
        return len(self.failure_examples) == 0 and len(self.success_examples) > 0

    @property
    def failed(self):
        return not self.passed

    @property
    def has_errors(self):
        return len(self.errors) > 0

    @property
    def passed_ratio(self):
        return len(self.success_examples) / (len(self.success_examples) + len(self.failure_examples))

    @classmethod
    def empty(cls):
        return EvaluationResult(list(), list(), list(), TestResultDetails.empty())

    def append(self, status: TestResultStatus, inputs, output, metadata, example: dict):
        self.details = self.details or TestResultDetails.empty()
        self.details.append(status, inputs, output, metadata)

        if status == TestResultStatus.PASSED:
            self.success_examples.append(example)
        elif status == TestResultStatus.FAILED:
            self.failure_examples.append(example)
        elif status == TestResultStatus.ERROR:
            self.errors.append(example)


class BaseEvaluator(ABC):
    """Base class for evaluators that define a way of detecting a LLM failure"""

    @abstractmethod
    def evaluate(self, model: BaseModel, dataset: Dataset):
        ...


class LLMBasedEvaluator(BaseEvaluator):
    _default_eval_prompt: str

    def __init__(self, eval_prompt=None, llm_temperature=0.1, llm_client: LLMClient = None, rng_seed: int = 1729):
        self.eval_prompt = eval_prompt or self._default_eval_prompt
        self.llm_temperature = llm_temperature
        self.llm_client = llm_client if llm_client is not None else get_default_client()
        self.rng_seed = rng_seed

    def _make_evaluate_prompt(self, model: BaseModel, input_vars, model_output, row_idx):
        return self.eval_prompt.format(
            model_name=model.name,
            model_description=model.description,
            input_vars=input_vars,
            model_output=model_output,
        )

    def _make_evaluate_functions(self, model: BaseModel, input_vars, model_output):
        return EVALUATE_MODEL_FUNCTIONS

    def _get_metadata(self, row_idx, *args, **kwargs) -> Dict[str, Any]:
        return dict()

    def evaluate(self, model: BaseModel, dataset: Dataset):
        model_outputs = model.predict(dataset).prediction

        evaluation_result = EvaluationResult.empty()
        for row_index, input_vars, model_output in zip(
            dataset.df.index,
            dataset.df.loc[:, model.feature_names].to_dict("records"),
            model_outputs,
        ):
            sample = {"input_vars": input_vars, "model_output": model_output}
            prompt = self._make_evaluate_prompt(model, input_vars, model_output, row_index)
            funcs = self._make_evaluate_functions(model, input_vars, model_output)
            metadata = self._get_metadata(row_index)
            try:
                out = self.llm_client.complete(
                    [{"role": "system", "content": prompt}],
                    tools=funcs,
                    tool_choice={"type": "function", "function": {"name": "evaluate_model"}},
                    temperature=self.llm_temperature,
                    caller_id=self.__class__.__name__,
                    seed=self.rng_seed,
                )
                if len(out.tool_calls) != 1 or "passed_test" not in out.tool_calls[0].function.arguments:
                    raise LLMGenerationError("Invalid function call arguments received")
            except LLMGenerationError as err:
                metadata["Reason"] = str(err)
                evaluation_result.append(
                    TestResultStatus.ERROR, input_vars, model_output, metadata, {"message": str(err), "sample": sample}
                )
                continue

            args = out.tool_calls[0].function.arguments
            metadata["Reason"] = args.get("reason")
            evaluation_result.append(
                TestResultStatus.PASSED if args["passed_test"] else TestResultStatus.FAILED,
                input_vars,
                model_output,
                metadata,
                {"input_vars": input_vars, "model_output": model_output, "reason": args.get("reason")},
            )

        return evaluation_result



from typing import Optional, Sequence, Union

import json
import logging
from enum import Enum

import numpy as np
import pandas as pd

from ..llm.client import get_default_client
from ..llm.client.base import LLMClient
from .prompts import (
    FIX_JSON_FORMAT_PROMPT,
    DistractingQuestionPrompt,
    QAGenerationPrompt,
    QuestionComplexificationPrompt,
)
from .testset import QATestset
from .vector_store import VectorStore

logger = logging.getLogger(__name__)


class DifficultyLevel(int, Enum):
    EASY = 1
    COMPLEX = 2
    DISTRACTING_ELEMENT = 3


class TestsetGenerator:
    """Testset generator for testing RAG models.

    Explore a given knowledge base and generate question/answer pairs to test the model.

    Each generated item contains the following field
    - question: a question about a part of the knowledge base
    - reference_answer: the expected answer according to the knowledge base
    - reference_context: relevant elements directly extracted from the knowledge base
    - difficulty_level: an indicator of how difficult the question is

    Parameters
    ----------
    knowledge_base: pd.DataFrame
        A dataframe containing the whole knowledge base.
    knowledge_base_columns: Sequence[str], optional
        The list of columns from the `knowledge_base` to consider. If not specified, all columns of the knowledge base
        dataframe will be concatenated to produce a single document.
        Example: if your knowledge base consists in FAQ data with columns "Q" and "A", we will format each row into a
        single document "Q: [question]\\nA: [answer]" to generate questions.
    language: str = "en"
        The language used to generate questions (e.g. "fr", "de", ...)
    model_name: str, optional
        Name of the model to be tested, to get more fitting questions.
    model_description: str, optional
        Description of the model to be tested.
    context_neighbors: int
        The maximum number of extracted element from the knowledge base to get a relevant context for question generation
    context_similarity_threshold: float = 0.2
        A similarity threshold to filter irrelevant element from the knowledge base during context creation
    context_window_length: int = 8192
        Context window length of the llm used in the `llm_client` of the generator
    embedding_fn: Callable = None
        Embedding function to build the knowledge base index.
    seed: int = None
    """

    def __init__(
        self,
        knowledge_base: pd.DataFrame,
        knowledge_base_columns: Sequence[str] = None,
        language: str = "en",
        model_name: str = None,
        model_description: str = None,
        context_neighbors: int = 4,
        context_similarity_threshold: float = 0.2,
        context_window_length: int = 8192,
        seed: int = None,
        include_examples: bool = True,
        embedding_model: str = "text-embedding-ada-002",
        llm_client: Optional[LLMClient] = None,
        llm_temperature: float = 0.5,
    ):
        self._knowledge_base = knowledge_base
        self._knowledge_base_columns = knowledge_base_columns
        self._language = language
        self._model_name = model_name
        self._model_description = model_description
        self._context_neighbors = context_neighbors
        self._context_similarity_threshold = context_similarity_threshold
        self._embedding_model = embedding_model
        self._context_window_length = context_window_length
        self._rng = np.random.default_rng(seed=seed)
        self._include_examples = include_examples
        self._vector_store_inst = None
        self._llm_client = llm_client or get_default_client()
        self._llm_temperature = llm_temperature

    @property
    def _vector_store(self):
        if self._vector_store_inst is None:
            logger.debug("Initializing vector store from knowledge base.")
            self._vector_store_inst = VectorStore.from_df(
                self._knowledge_base,
                lambda query: self._llm_client.embeddings(query, model=self._embedding_model),
                features=self._knowledge_base_columns,
            )
        return self._vector_store_inst

    def _get_generator_method(self, level: DifficultyLevel):
        mapping = {
            DifficultyLevel.EASY: self._generate_question_easy,
            DifficultyLevel.COMPLEX: self._generate_question_complex,
            DifficultyLevel.DISTRACTING_ELEMENT: self._generate_question_distracting_element,
        }

        try:
            return mapping[level]
        except KeyError:
            raise ValueError(f"Invalid difficulty level: {level}.")

    def _generate_question_easy(self, context: str) -> dict:
        messages = QAGenerationPrompt.create_messages(
            model_name=self._model_name,
            model_description=self._model_description,
            language=self._language,
            user_content=context,
        )

        generated_qa = self._llm_complete(messages=messages)
        generated_qa["difficulty"] = DifficultyLevel.EASY
        return generated_qa

    def _generate_question_complex(self, context: str) -> dict:
        generated_qa = self._generate_question_easy(context)

        messages = QuestionComplexificationPrompt.create_messages(
            model_name=self._model_name,
            model_description=self._model_description,
            language=self._language,
            user_content=(generated_qa["question"], context),
        )
        generated_qa["difficulty"] = DifficultyLevel.COMPLEX
        out = self._llm_complete(messages=messages)
        generated_qa["question"] = out["question"]
        return generated_qa

    def _generate_question_distracting_element(self, context: str) -> dict:
        generated_qa = self._generate_question_easy(context)

        distracting_context = self._rng.choice(self._vector_store.documents).content
        messages = DistractingQuestionPrompt.create_messages(
            model_name=self._model_name,
            model_description=self._model_description,
            language=self._language,
            user_content=(generated_qa["question"], generated_qa["answer"], distracting_context),
        )
        generated_qa["difficulty"] = DifficultyLevel.DISTRACTING_ELEMENT
        out = self._llm_complete(messages=messages)
        generated_qa["question"] = out["question"]
        return generated_qa

    def _get_random_document_group(self):
        seed_embedding = self._rng.choice(self._vector_store.embeddings)
        relevant_contexts = [
            context
            for (context, score) in self._vector_store.vector_similarity_search_with_score(
                seed_embedding, k=self._context_neighbors
            )
            if score < self._context_similarity_threshold
        ]

        return relevant_contexts

    def _prevent_context_window_overflow(self, prompt: str):
        # Prevent context overflow
        # general rule of thumbs to count tokens: 1 token ~ 4 characters
        # https://help.openai.com/en/articles/4936856-what-are-tokens-and-how-to-count-them
        return prompt[: self._context_window_length * 4]

    def _llm_complete(self, messages: Sequence[dict]) -> dict:
        try:
            out = self._llm_client.complete(
                messages=messages,
                temperature=self._llm_temperature,
                caller_id=self.__class__.__name__,
            )

            return json.loads(out.content, strict=False)
        except json.decoder.JSONDecodeError:
            logger.warning("JSON decoding error, trying to fix the JSON string.")
            return self._try_fix_json_message(out.content)

    def _try_fix_json_message(self, incorrect_json: str):
        out = self._llm_client.complete(
            messages=[
                {"role": "system", "content": FIX_JSON_FORMAT_PROMPT},
                {"role": "user", "content": incorrect_json},
            ],
            temperature=0,
            caller_id=self.__class__.__name__,
        )
        return json.loads(out.content)

    def generate_testset(
        self,
        num_questions: int = 10,
        difficulty: Union[DifficultyLevel, Sequence[DifficultyLevel]] = DifficultyLevel.EASY,
    ) -> QATestset:
        """Generates a testset from the knowledge base.

        Parameters
        ----------
        num_questions : int
            The number of question to generate for each difficulty level. By default 10.
        difficulty : Union[DifficultyLevel, Sequence[DifficultyLevel]]
            The difficulty level of the questions to generate. Can be 1 (:attr:`DifficultyLevel.EASY`), 2 (:attr:`DifficultyLevel.COMPLEX`),
            3 (:attr:`DifficultyLevel.DISTRACTING_ELEMENT`) or a list of these values. By default will use the easy level.

        Returns
        -------
        QATestset
            The generated test set.

        """
        if not isinstance(difficulty, Sequence):
            difficulty = [difficulty]

        generated_questions = []
        for level in difficulty:
            for idx in range(num_questions):
                logger.info(f"Generating question {idx + 1}/{num_questions} for difficulty level {str(level)}.")
                context_docs = self._get_random_document_group()
                context = QAGenerationPrompt.format_context(context_docs)

                generation_fn = self._get_generator_method(level)

                try:
                    generated_qa = generation_fn(context)
                except Exception as e:
                    logger.error(f"Encountered error in question generation: {e}. Skipping.")
                    continue

                generated_questions.append(
                    {
                        "question": generated_qa["question"],
                        "reference_answer": generated_qa["answer"],
                        "reference_context": context,
                        "difficulty_level": generated_qa["difficulty"],
                    }
                )

        return QATestset(pd.DataFrame(generated_questions))



import pandas as pd

from ..core.suite import Suite
from ..datasets.base import Dataset
from ..testing.tests.llm import test_llm_correctness


class QATestset:
    """A class to represent a testset for QA models."""

    def __init__(self, dataframe: pd.DataFrame):
        self._dataframe = dataframe

    def __len__(self):
        return len(self._dataframe)

    def to_pandas(self):
        """Return the testset as a pandas DataFrame."""
        return self._dataframe

    def to_dataset(self):
        return Dataset(self._dataframe, name="QA Testset", target=False, validation=False)

    def save(self, path):
        """Save the testset as a JSONL file.

        Parameters
        ----------
        path : str
            The path to the output JSONL file.
        """
        self._dataframe.to_json(path, orient="records", lines=True)

    @classmethod
    def load(cls, path):
        """Load a testset from a JSONL file.

        Parameters
        ----------
        path : str
            The path to the input JSONL file.
        """
        dataframe = pd.read_json(path, orient="records", lines=True)
        return cls(dataframe)

    def to_test_suite(self, name=None):
        """
        Convert the testset to a Giskard test suite.

        Parameters
        ----------
        name : str, optional
            The name of the test suite. If not provided, the name will be "Test suite generated from testset".

        Returns
        -------
        giskard.Suite
            The test suite.
        """
        suite_default_params = {"dataset": self.to_dataset()}
        name = name or "Test suite generated from testset"
        suite = Suite(name=name, default_params=suite_default_params)
        suite.add_test(test_llm_correctness, "TestsetCorrectnessTest", "TestsetCorrectnessTest")
        return suite

    def copy(self):
        """Return a copy of the testset."""
        return QATestset(self._dataframe.copy())



QA_GENERATION_SYSTEM_PROMPT_WITH_DESCRIPTION = """You are a powerful auditor, your role is to generate question & answer pair from a given list of context paragraphs.

The model you are auditing is the following:
- Model name: {model_name}
- Model description: {model_description}

Your question must be related to a provided context.  
Please respect the following rules to generate the question:
- The answer to the question should be found inside the provided context
- The question must be self-contained
- The question and answer must be in this language: {language}

The user will provide the context, consisting in multiple paragraphs delimited by dashes "------".
You will return the question and the precise answer to the question based exclusively on the provided context.
You must output a single JSON object with keys 'question' and 'answer'. Make sure you return a valid JSON object."""

QA_GENERATION_SYSTEM_PROMPT = """You are a powerful auditor, your role is to generate a question & answer pair from a given list of context paragraphs.

Your question must be related to a provided context.  
Please respect the following rules to generate the question:
- The answer to the question should be found inside the provided context
- The question must be self-contained
- The question and answer must be in this language: {language}

You will be provided the context, consisting in multiple paragraphs delimited by dashes "------".
You will return the question and the precise answer to the question based exclusively on the provided context.
Your output should be a single JSON object, with keys 'question' and 'answer'. Make sure you return a valid JSON object."""

QA_GENERATION_ASSISTANT_EXAMPLE = """{
    "question": "For which countries can I track my shipping?",
    "answer": "We ship to all 50 states in the US, as well as to Canada and Mexico. We offer tracking for all our shippings."
}"""

QA_GENERATION_CONTEXT_EXAMPLE = """What payment methods do you accept?

We accept a variety of payment methods to provide our customers with a convenient and secure shopping experience. You can make a purchase using major credit and debit cards, including Visa, Mastercard, American Express, and Discover. We also offer the option to pay with popular digital wallets such as PayPal and Google Pay. For added flexibility, you can choose to complete your order using bank transfers or wire transfers. Rest assured that we prioritize the security of your personal information and go the extra mile to ensure your transactions are processed safely.
------
\tWhat is your shipping policy?

We offer free shipping on all orders over $50. For orders below $50, we charge a flat rate of $5.99. We offer shipping services to customers residing in all 50\n states of the US, in addition to providing delivery options to Canada and Mexico.
------
\tHow can I track my order?

Tracking your order is a breeze! Once your purchase has been successfully confirmed and shipped, you will receive a confirmation email containing your tracking number. You can simply click on the link provided in the email or visit our website's order tracking page. Enter your tracking number, and you will be able to monitor the progress of your shipment in real-time. This way, you can stay updated on the estimated delivery date and ensure you're available to receive your package.
"""

FIX_JSON_FORMAT_PROMPT = """Fix the following json string so it contains a single valid json. Make sure to start and end with curly brackets."""


class QAGenerationPrompt:
    system_prompt_with_description = QA_GENERATION_SYSTEM_PROMPT_WITH_DESCRIPTION
    system_prompt_raw = QA_GENERATION_SYSTEM_PROMPT
    example_prompt = QA_GENERATION_CONTEXT_EXAMPLE
    example_answer = QA_GENERATION_ASSISTANT_EXAMPLE

    @classmethod
    def _format_system_prompt(cls, model_name, model_description, language):
        language = language or "en"
        if model_name is not None or model_description is not None:
            system_prompt = cls.system_prompt_with_description.format(
                model_name=model_name,
                model_description=model_description,
                language=language,
            )
        else:
            system_prompt = cls.system_prompt_raw.format(
                language=language,
            )
        system_message = {
            "role": "system",
            "content": system_prompt,
        }
        return system_message

    @classmethod
    def _format_example_prompt(cls, examples):
        if examples is not None:
            return examples
        elif cls.example_prompt is not None:
            examples = []
            if cls.example_prompt is not None:
                examples.append({"role": "user", "content": cls.example_prompt})
            if cls.example_prompt is not None:
                examples.append({"role": "assistant", "content": cls.example_answer})
            return examples
        return []

    @classmethod
    def format_context(cls, contexts):
        return "\n------\n".join(["", *[doc.content for doc in contexts], ""])

    @classmethod
    def create_messages(
        cls,
        model_name=None,
        model_description=None,
        language=None,
        add_examples=False,
        examples=None,
        user_content=None,
    ):
        messages = [cls._format_system_prompt(model_name, model_description, language)]
        if add_examples:
            messages.extend(cls._format_example_prompt(examples))

        if user_content is not None:
            messages.append({"role": "user", "content": user_content})

        return messages


COMPLEXIFICATION_SYSTEM_PROMPT_WITH_DESCRIPTION = """You are an expert at writing questions. 
Your task is to re-write questions that will be used to evaluate the following model:
- Model name: {model_name}
- Model description: {model_description}  

Respect the following rules to reformulate the question:
- The re-written question should not be longer than the original question by up to 10 to 15 words. 
- The re-written question should be more elaborated than the original, use elements from the context to enrich the questions. 
- The re-written question should be more difficult to handle for AI models but it must be understood and answerable by humans.
- Add one or more constraints / conditions to the question.
- The re-written question must be in this language: {language}

You will be provided the question delimited by <question></question> tags.
You will also be provided a relevant context which contain the answer to the question, delimited by <context></context> tags. It consists in multiple paragraphs delimited by dashes "------".
You will return the reformulated question as a single JSON object, with the key 'question'. Make sure you return a valid JSON object.
"""

COMPLEXIFICATION_SYSTEM_PROMPT = """You are an expert at writing questions. 
Your task is to re-write questions that will be used to evaluate a language model.

Respect the following rules to reformulate the question:
- The re-written question should not be longer than the original question by up to 10 to 15 words. 
- The re-written question should be more elaborated than the original, use elements from the context to enrich the questions. 
- The re-written question should be more difficult to handle for AI models but it must be understood and answerable by humans.
- Add one or more constraints / conditions to the question.
- The re-written question must be in this language: {language}

You will be provided the question delimited with <question></question> tags.
You will also be provided a relevant context which contain the answer to the question, delimited with <context></context> tags. It consists in multiple paragraphs delimited by dashes "------".
You will return the reformulated question as a single JSON object, with the key 'question'. Make sure you return a valid JSON object.
"""

COMPLEXIFICATION_PROMPT_EXAMPLE = """<question>
For which countries can I track my shipping?
</question>

<context>
What payment methods do you accept?

\tWe accept a variety of payment methods to provide our customers with a convenient and secure shopping experience. You can make a purchase using major credit and debit cards, including Visa, Mastercard, American Express, and Discover. We also offer the option to pay with popular digital wallets such as PayPal and Google Pay. For added flexibility, you can choose to complete your order using bank transfers or wire transfers. Rest assured that we prioritize the security of your personal information and go the extra mile to ensure your transactions are processed safely.
------
\tWhat is your shipping policy?

We offer free shipping on all orders over $50. For orders below $50, we charge a flat rate of $5.99. We offer shipping services to customers residing in all 50\n states of the US, in addition to providing delivery options to Canada and Mexico.
------
\tHow can I track my order?

Tracking your order is a breeze! Once your purchase has been successfully confirmed and shipped, you will receive a confirmation email containing your tracking number. You can simply click on the link provided in the email or visit our website's order tracking page. Enter your tracking number, and you will be able to monitor the progress of your shipment in real-time. This way, you can stay updated on the estimated delivery date and ensure you're available to receive your package.
<context>
"""

COMPLEXIFICATION_ANSWER_EXAMPLE = """{
    "question": "Can you provide my a list of the countries from which I can follow the advancement of the delivery of my shipping?"
}"""


class QuestionComplexificationPrompt(QAGenerationPrompt):
    system_prompt_with_description = COMPLEXIFICATION_SYSTEM_PROMPT_WITH_DESCRIPTION
    system_prompt_raw = COMPLEXIFICATION_SYSTEM_PROMPT
    example_prompt = COMPLEXIFICATION_PROMPT_EXAMPLE
    example_answer = COMPLEXIFICATION_ANSWER_EXAMPLE

    @classmethod
    def format_user_content(cls, question, context):
        context_string = f"<question>\n{question}\n</question>\n<context>\n{context}\n</context>"
        return context_string

    @classmethod
    def create_messages(cls, **kwargs):
        kwargs["user_content"] = cls.format_user_content(*kwargs["user_content"])
        return super().create_messages(**kwargs)


DISTRACTING_QUESTION_SYSTEM_PROMPT = """You are an expert at rewriting question.
Your task is to re-write questions that will be used to evaluate a language model.

Your task is to complexify questions given a provided context. 
Please respect the following rules to generate the question:
- The new question must include a condition or constraint based on the provided context. 
- The new question must have the same answer as the original question.
- The question must be plausible according to the context and the model description.
- The question must be self-contained and understandable by humans. 
- The question must be in this language: {language}

You will be provided the question and its answer delimited with <question></question> and <answer></answer> tags.
You will also be provided a context paragraph delimited with <context></context> tags.
You will return the reformulated question as a single JSON object, with the key 'question'. Make sure you return a valid JSON object.
"""

DISTRACTING_QUESTION_SYSTEM_PROMPT_WITH_DESCRIPTION = """You are an expert at rewriting questions.
Your task is to re-write questions that will be used to evaluate the following model:
- Model name: {model_name}
- Model description: {model_description}  

Your task is to complexify questions given a provided context. 
Please respect the following rules to generate the question:
- The new question must include a condition or constraint based on the provided context. 
- The original question direction should be preserved.
- The question must be plausible according to the context and the model description.
- The question must be self-contained and understandable by humans. 
- The question must be in this language: {language}

You will be provided the question delimited with <question></question> tags.
You will also be provided a context paragraph delimited with <context></context> tags.
You will return the reformulated question as a single JSON object, with the key 'question'. Make sure you return a valid JSON object.
"""

DISTRACTING_QUESTION_USER_INPUT = """<question>
{question}
</question>
<answer>
{answer}
</answer>
<context>
{context}
</context>"""

DISCTRACTING_QUESTION_PROMPT_EXAMPLE = DISTRACTING_QUESTION_USER_INPUT.format(
    question="What job offer do you have for engineering student?",
    answer="We have plenty of different jobs for engineering student depending on your speciality: mechanical engineer, data scientist, electronic designer and many more.",
    context="Sometimes employers assume being accessible and inclusive only means providing physical access like ramps, accessible bathrooms and automatic opening doors. However, there are many other important ways to demonstrate that you welcome and want to attract a diverse workforce including people with disability.",
)

DISCTRACTING_QUESTION_ANSWER_EXAMPLE = """{
    "question": "Do you have any job opening suitable for engineering students with a disability? "
}"""


class DistractingQuestionPrompt(QuestionComplexificationPrompt):
    system_prompt_with_description = DISTRACTING_QUESTION_SYSTEM_PROMPT_WITH_DESCRIPTION
    system_prompt_raw = DISTRACTING_QUESTION_SYSTEM_PROMPT
    example_prompt = DISCTRACTING_QUESTION_PROMPT_EXAMPLE
    example_answer = DISCTRACTING_QUESTION_ANSWER_EXAMPLE

    @classmethod
    def format_user_content(cls, question, answer, context):
        return DISTRACTING_QUESTION_USER_INPUT.format(question=question, answer=answer, context=context)



from typing import Any, Dict, List, Optional, Sequence

from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np

from .logger import LLMLogger


@dataclass
class LLMFunctionCall:
    name: str
    arguments: Any


@dataclass
class LLMToolCall:
    id: str
    type: str
    function: LLMFunctionCall


@dataclass
class LLMMessage:
    role: str
    content: Optional[str] = None
    function_call: Optional[LLMFunctionCall] = None
    tool_calls: Optional[List[LLMToolCall]] = None

    @staticmethod
    def create_message(role: str, content: str):
        return LLMMessage(role=role, content=content, function_call=None, tool_calls=None)


class LLMClient(ABC):
    @property
    @abstractmethod
    def logger(self) -> LLMLogger:
        ...

    @abstractmethod
    def complete(
        self,
        messages: Sequence[LLMMessage],
        functions=None,
        temperature=0.5,
        max_tokens=None,
        function_call: Optional[Dict] = None,
        caller_id: Optional[str] = None,
        tools=None,
        tool_choice=None,
        seed: Optional[int] = None,
    ) -> LLMMessage:
        ...

    @abstractmethod
    def embeddings(self, text) -> np.ndarray:
        ...
